"""
    Fastberry Main (Core-Methods)
"""

from starlette.middleware.base import BaseHTTPMiddleware as BaseMiddleware
from strawberry.extensions import Extension as BaseExtension
from strawberry.permission import BasePermission

from . import config

# from .config import Settings
from .config import Settings as Fastberry
from .config.extras.gql import GQL

# from .graphql import BaseType
from .schema import Schema
from .types import Model

Text = Model.text
JSON = Model.json



import dataclasses as dc
import functools
import typing
from collections import OrderedDict

from .scalars import (
    JSON,
    SQLALCHEMY_ACTIVE,
    SQLALCHEMY_BASE,
    STRAWBERRY_CORE,
    STRAWBERRY_FIELDS,
    TEXT,
    AnnotationConfig,
    Manager,
    Scalars,
    TableConfig,
    TypeConfig,
    get_module_name,
)
from .sqlalchemy import create_sqlalchemy_model

CUSTOM_SCALARS = [TEXT, JSON]


class Globals:
    """API Globals"""

    @staticmethod
    def set(key, value):
        globals().update({key: value})

    @staticmethod
    def get(key):
        return globals().get(key)


class BaseClass:
    """API Base Class"""

    _id: typing.Optional[str] = None
    id: typing.Optional[STRAWBERRY_FIELDS.ID] = None


def get_field_annotation_core(single_annotation: typing.Any, found_args: tuple):
    """Field Core Annotation"""
    core_arg = None
    has_children = []

    # Core Arg
    if found_args:
        core_arg = found_args[0]
    else:
        core_arg = single_annotation

    # Core Arg Continued . . .
    has_children = typing.get_args(core_arg)
    if len(has_children) > 0:
        core_arg = has_children[0]
        if isinstance(core_arg, str):
            core_arg = typing.ForwardRef(core_arg)
    return core_arg


def get_annotations_args(single_annotation: typing.Any):
    """Get: Annotations-Data"""
    found_args = typing.get_args(single_annotation)
    real_arg = get_field_annotation_core(single_annotation, found_args)
    is_custom_type = isinstance(real_arg, typing.ForwardRef)
    is_optional = type(None) in found_args
    scalar_arg = None

    # print(real_arg)
    if real_arg in CUSTOM_SCALARS:
        fake_arg = Scalars.get(real_arg)
        scalar_arg = real_arg
        real_arg = fake_arg.gql
    else:
        scalar_arg = real_arg

    # IS -> List <YES or NO>
    if found_args and len(typing.get_args(found_args[0])) > 0:
        is_list = True
    else:
        is_list = False

    # IS -> Step 2 ...
    if len(found_args) == 1:
        if found_args[0] in CUSTOM_SCALARS:
            is_list = True

    # IS -> Step 3 ...
    if is_list:
        core_arg = typing.List[real_arg]
    else:
        core_arg = real_arg

    # Custom Annotations
    return AnnotationConfig(
        original=single_annotation,
        real=real_arg,
        core=core_arg,
        types=Scalars.get(scalar_arg),
        optional=is_optional,
        is_list=is_list,
        is_dict=scalar_arg == JSON,
        is_custom_type=is_custom_type,
    )


def get_function_fields(model: object):
    """Get Object Fields"""
    __fields = [i for i in dir(model) if not i.startswith("__")]
    return {key: getattr(model, key) for key in __fields}


class Model:
    sqlalchemy_base = SQLALCHEMY_BASE
    load = staticmethod(Manager.load)
    json = JSON
    text = TEXT

    def __init__(self, sql=None, mongo=None):
        """Create Custon <Type>"""

        if SQLALCHEMY_ACTIVE:
            SQLBase = (sql,)
        else:
            SQLBase = (object,)

        # Create Custon <Type>
        @staticmethod
        def custom_type(
            original_object: object = None,
            *,
            table_name: str = None,
            primary_key: list = None,
            required: list = None,
            index: list = None,
            unique: list = None,
            unique_together: list = None,
            many_to_many: list = None,
            is_sql: bool = False,
            is_mongo: bool = False,
        ):
            """DECORATOR-OPTIONAL-ARGUMENT: https://pybit.es/articles/decorator-optional-argument/"""
            primary_key = primary_key or []
            required = required or []
            index = index or []
            unique = unique or []
            unique_together = unique_together or []
            many_to_many = many_to_many or []

            # Starting Wrapper. . .
            if original_object is None:
                return functools.partial(
                    custom_type,
                    table_name=table_name,
                    primary_key=primary_key,
                    required=required,
                    index=index,
                    unique=unique,
                    unique_together=unique_together,
                    many_to_many=many_to_many,
                    is_sql=is_sql,
                    is_mongo=is_mongo,
                )

            # Class [ INFO ]
            class_core_annotations = OrderedDict()
            class_related_fields = OrderedDict()
            class_default_values = OrderedDict()
            class_fields_custom_annotations = OrderedDict()

            # Class [ Utils ]
            util_dataclass_fields = OrderedDict()
            util_default_field_is_auto = False

            # Class [ CUSTOM ]
            custom_class = type(
                original_object.__name__,
                (BaseClass,),
                {},
            )
            custom_class.__annotations__ = {
                **BaseClass.__annotations__,
                **original_object.__annotations__,
            }
            custom_dataclass = dc.dataclass(original_object)

            # Create Default Fields
            for field in dc.fields(custom_dataclass):
                if not isinstance(field.default_factory, dc._MISSING_TYPE):
                    default_value = dc.field(default_factory=field.default_factory)
                elif not isinstance(field.default, dc._MISSING_TYPE):
                    default_value = dc.field(default=field.default)
                else:
                    default_value = dc.field(default=None)
                    util_default_field_is_auto = True
                # Register the (Default Value)
                util_dataclass_fields[field.name] = default_value

            # Field & Annotation
            for field_name, annotation in original_object.__annotations__.items():
                current_annotation = get_annotations_args(annotation)
                # Create the (Default Value)
                default_value = util_dataclass_fields[field_name]
                if util_default_field_is_auto and current_annotation.is_dict:
                    default_value = dc.field(default_factory=dict)
                elif util_default_field_is_auto and current_annotation.is_list:
                    default_value = dc.field(default_factory=list)
                # Default Value
                class_default_values[field_name] = default_value
                # Set the (Default Value)
                setattr(custom_class, field_name, default_value)
                custom_class.__annotations__[field_name] = typing.Optional[
                    current_annotation.core
                ]
                # Register the Annotation and Field for Database
                class_fields_custom_annotations[field_name] = current_annotation
                class_core_annotations[field_name] = current_annotation.core
                # IS Relationships
                if current_annotation.is_custom_type:
                    class_related_fields[field_name] = current_annotation.real

            for field_name, field_function in get_function_fields(
                original_object
            ).items():
                if callable(field_function):
                    # Get the (Original Function)
                    func_annotation = field_function.__annotations__.get("return")
                    current_annotation = get_annotations_args(func_annotation)
                    # Register the Annotation and Field for Database
                    class_fields_custom_annotations[field_name] = current_annotation
                    class_core_annotations[field_name] = current_annotation.core
                    # Set the (Custom Function)
                    if STRAWBERRY_CORE:
                        field_function = STRAWBERRY_CORE.field(field_function)
                    setattr(custom_class, field_name, field_function)
                    # IS Relationships
                    if current_annotation.is_custom_type:
                        class_related_fields[field_name] = current_annotation.real

            # Strawberry
            if STRAWBERRY_CORE:
                custom_class = STRAWBERRY_CORE.type(custom_class)

            # Info
            info_class_name = original_object.__name__
            info_class_module_path = original_object.__module__
            info_class_module_name = get_module_name(info_class_module_path)

            # Info Table
            info_class_table_name = (
                f"{info_class_module_name.lower()}_{info_class_name.lower()}"
            )
            info_class_table_uri = (
                f"{info_class_module_name.lower()}.{info_class_name.lower()}"
            )
            if table_name:
                info_class_table_name = table_name

            # Table Config
            table_config = TableConfig(
                required=lambda key: (key in required),
                primary_key=lambda key: (key in primary_key),
                index=lambda key: (key in index),
                unique=lambda key: (key in unique),
                unique_together=unique_together,
                many_to_many=many_to_many,
            )

            # Type Config
            type_config = TypeConfig(
                custom_class=custom_class,
                name=info_class_name,
                module=info_class_module_name,
                module_path=info_class_module_path,
                table_name=info_class_table_name,
                table_uri=info_class_table_uri,
                core_fields=class_core_annotations,
                related_fields=class_related_fields,
                default_values=class_default_values,
                custom_annotations=class_fields_custom_annotations,
                table_config=table_config,
                sql=is_sql,
                mongo=is_mongo,
            )

            # Attach <Objects>
            custom_class._lazy_object = False
            custom_class._relationships = None
            if is_mongo:
                custom_class.objects = mongo[info_class_table_name]
            elif is_sql:
                sql_obj = create_sqlalchemy_model(SQLBase, type_config)
                custom_class._lazy_object = sql_obj.lazy
                custom_class._relationships = sql_obj._relationships
                if sql_obj.lazy:
                    custom_class.objects = functools.partial(
                        sql_obj.objects, custom_class
                    )
                else:
                    custom_class.objects = sql_obj.objects
            else:
                custom_class.objects = None

            # Attach <Meta>
            custom_class.__meta__ = type_config

            # Register and Return
            Manager.register_model(name=type_config.name, model=type_config)
            return custom_class

        # Self Definitions
        self.type = custom_type
        self.sql = functools.partial(custom_type, is_sql=True)
        self.mongo = functools.partial(custom_type, is_mongo=True)



"""C.R.U.D Base

classes:
    GQL: The base for GraphQL (Query & Mutation) methods.

methods:
    is_crud: Check if value is a subclass ot <CRUD>.
"""

import strawberry
from strawberry.types import Info

from ...utils.objects import Singleton, get_attr
from .. import core
from ..imports import search_method


def api_name(self):
    """API Name Maker"""

    if self.name:
        return f"{self.name}API"
    return "API"


def is_crud(crud):
    """Check if item is subclass of <GQL>"""

    its_crud = False
    try:
        if not crud == GQL:
            its_crud = issubclass(crud, GQL)
    except TypeError:
        pass
    return its_crud


class GQL(Singleton):
    """CRUD (GraphQL) Singleton"""

    def init(self):
        """__init__ Only Runs Once Per Class."""

        settings = core.Settings()

        permission_classes = []
        for perm in settings.base.__dict__.get("permissions") or []:
            permission = search_method(perm)
            if permission:
                permission_classes.append(permission)

        try:
            self.name = self.schema.__name__
        except AttributeError:
            self.name = None

        app_prefix = get_attr(self, "prefix")
        all_mutation = get_attr(self, "Mutation")
        all_query = get_attr(self, "Query")

        dict_names = {"Query": [], "Mutation": []}
        dict_mutation = {}
        dict_query = {}

        # Naming
        def func_name(key, group=self.name, prefix=app_prefix):
            """Create the function's name."""
            if group:
                schema_name = group.lower()
                return_value = f"{schema_name}_{key.lower()}"
            else:
                return_value = key.lower()
            if prefix:
                return_value = f"{prefix}_{return_value}"
            return return_value

        # Mutation
        if hasattr(all_mutation, "keys"):
            for key in all_mutation.keys:
                resolver_name = func_name(key)
                resolver_func = get_attr(all_mutation, key)
                resolver_func.__annotations__["info"] = Info
                dict_names["Mutation"].append(resolver_name)
                dict_mutation[resolver_name] = strawberry.field(
                    resolver=resolver_func,
                    description=(resolver_func.__doc__ or "").strip(),
                    permission_classes=permission_classes,
                )

        # Query
        if hasattr(all_query, "keys"):
            for key in all_query.keys:
                resolver_name = func_name(key)
                resolver_func = get_attr(all_query, key)
                resolver_func.__annotations__["info"] = Info
                dict_names["Query"].append(resolver_name)
                dict_query[resolver_name] = strawberry.field(
                    resolver=resolver_func,
                    description=(resolver_func.__doc__ or "").strip(),
                    permission_classes=permission_classes,
                )

        # Names
        self.Operations = dict_names

        # Write
        self.Mutation = None
        if dict_mutation:
            self.Mutation = strawberry.type(
                type(
                    api_name(self),
                    (object,),
                    dict_mutation,
                )
            )

        # Read
        self.Query = None
        if dict_query:
            self.Query = strawberry.type(
                type(
                    api_name(self),
                    (object,),
                    dict_query,
                )
            )



"""[API Types] - Map DataClass(Fields) to SQL(Fields) and GQL(Types)."""

import dataclasses
import datetime
import decimal
import types
import typing
from collections import OrderedDict

# SQLAlchemy
try:
    import sqlalchemy
    from sqlalchemy.ext.declarative import declarative_base

    SQLALCHEMY_ACTIVE = True
    SQLALCHEMY_CORE = sqlalchemy
except ImportError:
    SQLALCHEMY_ACTIVE = False
    SQLALCHEMY_CORE = None

# Strawberry
try:
    import strawberry

    STRAWBERRY_ACTIVE = True
    STRAWBERRY_CORE = strawberry
except ImportError:
    STRAWBERRY_ACTIVE = False
    STRAWBERRY_CORE = None


# Custom Typing
TEXT = typing.TypeVar("TEXT", str, None)
JSON = typing.TypeVar("JSON", object, dict, list, None)


@dataclasses.dataclass(frozen=True)
class AnnotationConfig:
    """Annotation Setup"""

    original: typing.Any
    real: typing.Any
    core: typing.Any
    types: typing.Any
    optional: bool = False
    is_list: bool = False
    is_dict: bool = False
    is_custom_type: bool = False


@dataclasses.dataclass
class APIType:
    """API-Field-Type"""

    name: str
    python: typing.Any
    sql: typing.Any
    gql: typing.Any


@dataclasses.dataclass(frozen=True)
class TableConfig:
    """Table Setup"""

    # table_name: str = None
    required: list = None
    primary_key: list = None
    index: list = None
    unique: list = None
    unique_together: list = None
    many_to_many: dict = None
    # is_sql: bool = False
    # is_mongo: bool = False


@dataclasses.dataclass(frozen=True)
class TypeConfig:
    """Type Setup"""

    custom_class: typing.Any
    name: str  # class_name
    module: str  # class_module
    module_path: str
    core_fields: dict
    related_fields: dict
    default_values: dict
    custom_annotations: dict
    table_config: TableConfig
    table_name: str  # for Databases
    table_uri: str  # for GLOBALS
    mongo: bool = False
    sql: bool = False
    is_super_class: bool = True


# SQLAlchemy-BASE
if SQLALCHEMY_ACTIVE:
    SQLALCHEMY_BASE = declarative_base()
else:
    SQLALCHEMY_BASE = object


class Singleton:
    """Create a Singleton"""

    def __new__(cls, *args, **kwargs):
        it_id = "__it__"
        it = cls.__dict__.get(it_id, None)
        if it is not None:
            return it
        it = object.__new__(cls)
        setattr(cls, it_id, it)
        it.init(*args, **kwargs)
        return it

    def init(self, *args, **kwargs):
        """Class __init__ Replacement"""


# SQL Relationships
class Globals:
    """API Globals"""

    @staticmethod
    def set(key, value):
        globals().update({key: value})

    @staticmethod
    def get(key):
        return globals().get(key)


class SQLManager(Singleton):
    def init(self, *args, **kwargs):
        """Class __init__ Replacement"""
        self.active = SQLALCHEMY_ACTIVE
        self.base = SQLALCHEMY_BASE
        self.models = OrderedDict()
        self.globals = Globals

    def register_model(self, name: str, model: TypeConfig):
        if name in self.models.keys():
            raise Exception(
                f"Duplicate Type: {name}\nExisting Model: {model.module_path}"
            )
        self.models[name] = model
        self.globals.set(name, model.custom_class)

    def find_model(self, name: str):
        return self.globals.get(name)

    @staticmethod
    def load(*models):
        for model in models:
            if model._lazy_object:
                model.objects()


def sql_easy_relationship(
    model_name: str = None,
    table_name: str = None,
    nullable: bool = True,
    unique: bool = False,
    index: bool = False,
):
    """Create a relationship in SQLAlchemy"""

    if SQLALCHEMY_ACTIVE:
        return sqlalchemy.Column(
            f"""{model_name}_id""",
            sqlalchemy.Integer,
            sqlalchemy.ForeignKey(f"""{table_name}._id"""),
            nullable=nullable,
            unique=unique,
            index=index,
        )
    return None


def sql_many_to_many(
    sql_base: typing.Any = None,
    table_left: str = None,
    table_right: str = None,
):
    """Create a Many-To-Many relationship in SQLAlchemy"""

    if SQLALCHEMY_ACTIVE:
        return sqlalchemy.Table(
            f"""association_{table_left}_and_{table_right}""",
            sql_base.metadata,
            sqlalchemy.Column(
                "left_id",
                sqlalchemy.String(length=255),
                primary_key=True,
            ),
            sqlalchemy.Column(
                "right_id",
                sqlalchemy.String(length=255),
                primary_key=True,
            ),
            sqlalchemy.UniqueConstraint("left_id", "right_id"),
        )
    return None


# CORE Fields
ALL_FIELDS = [
    "String",
    "Text",
    "Integer",
    "Float",
    "Boolean",
    "Date",
    "DateTime",
    "Time",
    "Decimal",
    "JSON",
    "ForeignKey",
    "UniqueTogether",
]

# Init Fields
SQLALCHEMY_FIELDS = types.SimpleNamespace(**{key: key for key in ALL_FIELDS})
STRAWBERRY_FIELDS = types.SimpleNamespace(**{key: key for key in ["ID", "JSON"]})

SQLALCHEMY_COLUMN = None
SQLALCHEMY_UNIQUE = None

STRAWBERRY_FIELDS.ID = str
STRAWBERRY_FIELDS.JSON = dict

# Map SQL
if SQLALCHEMY_ACTIVE:
    # Core
    SQLALCHEMY_FIELDS.Base = SQLALCHEMY_BASE
    SQLALCHEMY_FIELDS.Table = sqlalchemy.Table
    SQLALCHEMY_FIELDS.Column = sqlalchemy.Column
    SQLALCHEMY_FIELDS.ForeignKey = sqlalchemy.ForeignKey
    SQLALCHEMY_FIELDS.UniqueTogether = sqlalchemy.UniqueConstraint
    # Fields
    SQLALCHEMY_FIELDS.String = sqlalchemy.String(length=255)
    SQLALCHEMY_FIELDS.Text = sqlalchemy.Text
    SQLALCHEMY_FIELDS.Integer = sqlalchemy.Integer
    SQLALCHEMY_FIELDS.Float = sqlalchemy.Float
    SQLALCHEMY_FIELDS.Boolean = sqlalchemy.Boolean
    SQLALCHEMY_FIELDS.Date = sqlalchemy.Date
    SQLALCHEMY_FIELDS.DateTime = sqlalchemy.DateTime
    SQLALCHEMY_FIELDS.Time = sqlalchemy.Time
    SQLALCHEMY_FIELDS.Decimal = sqlalchemy.DECIMAL
    SQLALCHEMY_FIELDS.JSON = sqlalchemy.JSON


# Map GQL
if STRAWBERRY_ACTIVE:
    STRAWBERRY_FIELDS.ID = strawberry.ID
    STRAWBERRY_FIELDS.JSON = strawberry.scalars.JSON

# Core Init
CORE = dict()

# Core Values
CORE[TEXT] = APIType(
    name="Text",
    python=TEXT,
    sql=SQLALCHEMY_FIELDS.Text,
    gql=str,
)

CORE[JSON] = APIType(
    name="JSON",
    python=JSON,
    sql=SQLALCHEMY_FIELDS.JSON,
    gql=STRAWBERRY_FIELDS.JSON,
)

CORE[str] = APIType(
    name="String",
    python=str,
    sql=SQLALCHEMY_FIELDS.String,
    gql=str,
)

CORE[int] = APIType(
    name="Integer",
    python=int,
    sql=SQLALCHEMY_FIELDS.Integer,
    gql=int,
)

CORE[float] = APIType(
    name="Float",
    python=float,
    sql=SQLALCHEMY_FIELDS.Float,
    gql=float,
)

CORE[bool] = APIType(
    name="Boolean",
    python=bool,
    sql=SQLALCHEMY_FIELDS.Boolean,
    gql=bool,
)

CORE[datetime.date] = APIType(
    name="Date",
    python=datetime.date,
    sql=SQLALCHEMY_FIELDS.Date,
    gql=datetime.date,
)

CORE[datetime.datetime] = APIType(
    name="DateTime",
    python=datetime.datetime,
    sql=SQLALCHEMY_FIELDS.DateTime,
    gql=datetime.datetime,
)

CORE[datetime.time] = APIType(
    name="Time",
    python=datetime.time,
    sql=SQLALCHEMY_FIELDS.Time,
    gql=datetime.time,
)

CORE[decimal.Decimal] = APIType(
    name="Decimal",
    python=decimal.Decimal,
    sql=SQLALCHEMY_FIELDS.Decimal,
    gql=decimal.Decimal,
)

CORE_ID = APIType(
    name="ID",
    python=str,
    sql=str,
    gql=STRAWBERRY_FIELDS.ID,
)


def get_module_name(name: str):
    """Get: Class-Module's Name"""
    parts = name.split(".")
    if parts[0] == "apps" and len(parts) > 1:
        parts.pop(0)
    module_name = parts[0]
    if module_name == "__main__":
        module_name = None
    return module_name


# SCALARS
Scalars = CORE.copy()

# MANAGER
Manager = SQLManager()



""" Fastberry Core

    import pathlib

    # Build paths inside the project like this: BASE_DIR / "subdir".
    BASE_DIR = pathlib.Path(__file__).resolve().parents[1]

    # Or like this: settings.base_dir / "subdir".
    settings = Settings(base_dir=BASE_DIR)

classes:
    Settings: Singleton.
"""

import os
import pathlib
import sys
import types

import click
from fastapi.routing import APIRouter

from ..utils.objects import Singleton, get_attr, get_fields
from .definitions import load_docs, load_env, load_mode, load_yaml
from .extras.click import command_collection, is_click
from .extras.strawberry import process_strawberry_crud
from .imports import import_modules, search_method


class Settings(Singleton):
    """
    # Use it like this in your scripts.
    settings = Settings()
    """

    def init(self, base_dir: pathlib.Path = None, is_cli: bool = False):
        """__init__ Only Runs Once Per Settings Project."""

        if not base_dir:
            raise Exception("Missing: Settings(base_dir=BASE_DIR)")

        # Core
        core = load_yaml()
        env_mode = load_mode().get("mode", "development")
        env = load_env(env_mode)

        # BASE_DIR
        self.base_dir = base_dir

        # Setup
        self.base = core
        self.env = env
        self.docs = load_docs()
        self.mode = env_mode
        self.debug = env.debug
        self.secret_key = env.secret_key

        # Router
        router = APIRouter()
        self.router = router

        # Keys
        self.keys = [
            "apps",
            "base",
            "base_dir",
            "cli",
            "debug",
            "docs",
            "env",
            "extensions",
            "middleware",
            "mode",
            "router",
            "secret_key",
            "pagination",
        ]

        # Include BASE_DIR / "apps"
        if base_dir:
            sys.path.insert(0, os.path.join(base_dir, "apps"))

        # Merge Apps (Installed + Development)
        installed_apps = core.__dict__.get("installed_apps")
        installed_apps = installed_apps or []
        if env_mode == "development":
            development_apps = core.__dict__.get("development_apps", [])
            development_apps = development_apps or []
            if development_apps:
                installed_apps.extend(development_apps)

        # Auto-Load Modules
        modules = {}
        all_clis = set()
        gql_schema = {"Query": [], "Mutation": []}
        gql_field_names = {"query": [], "mutation": []}
        gql_paths = set()
        app_models = {}

        # IF Found Apps
        if installed_apps:
            modules = import_modules(installed_apps)
            for app_name, app_module in modules.items():
                if app_name.endswith(".types"):
                    for possible_type in get_fields(app_module):
                        current_type = get_attr(app_module, possible_type)
                        if hasattr(current_type, "__meta__"):
                            app_models[current_type.__meta__.table_uri] = current_type
                    # router = get_attr(app_module, "router")
                elif app_name.endswith(".graphql"):
                    # Load CRUD(Query, Mutation)
                    gql_types = process_strawberry_crud(app_module)
                    gql_schema["Query"].extend(gql_types["Query"])
                    gql_schema["Mutation"].extend(gql_types["Mutation"])
                    gql_field_names["query"].extend(
                        gql_types["Operations"]["Query"])
                    gql_field_names["mutation"].extend(
                        gql_types["Operations"]["Mutation"]
                    )
                    # Load Operations Folders <Dirs>
                    gql_path = pathlib.Path(app_module.__file__).parents[0]
                    if gql_path.name == "crud":
                        gql_path = gql_path.parents[0]
                    gql_paths.add(gql_path)
                elif app_name.endswith(".router"):
                    router = get_attr(app_module, "router")
                    # Load Routers
                    if isinstance(router, APIRouter):
                        self.router.include_router(router)
                elif app_name.endswith(".commands") and is_cli:
                    # Load Commands
                    cli = get_attr(app_module, "cli")
                    if is_click(cli):
                        all_clis.add(cli)

        # All - CLI(s)
        all_clis = list(all_clis)

        # Middleware
        installed_middleware = []
        for middleware in core.__dict__.get("middleware") or []:
            current_method = search_method(middleware)
            if current_method:
                installed_middleware.append(current_method)

        def add_middleware(app):
            for middleware in reversed(installed_middleware):
                app.add_middleware(middleware)
            return app

        # Extensions
        graphql_extensions = []
        for extension in core.__dict__.get("extensions") or []:
            current = search_method(extension)
            if current:
                graphql_extensions.append(current)

        # EVENTS
        events_extensions = {
            "startup": [],
            "shutdown": [],
        }
        for func in core.__dict__.get("on_startup") or []:
            current = search_method(func)
            if current:
                events_extensions["startup"].append(current)
        for func in core.__dict__.get("on_shutdown") or []:
            current = search_method(func)
            if current:
                events_extensions["shutdown"].append(current)

        # Pagination
        items_per_page = core.querying.get("items_per_page", 50)

        def pagination(page: int = 0, limit: int = 100):
            if not page:
                page = 1
            if not limit:
                limit = items_per_page
            limit = min(items_per_page, limit)
            page = min(1, page)
            return types.SimpleNamespace(page=page, limit=limit)

        # Easy-to-Use Apps
        super_api = types.SimpleNamespace(
            **dict(
                modules=modules,
                schema=types.SimpleNamespace(**gql_schema),
                pagination=pagination,
                paths=gql_paths,
                operations=types.SimpleNamespace(**gql_field_names),
            )
        )

        # Load Lazy SQL Objects
        for current_type in app_models.values():
            if current_type._lazy_object:
                current_type.objects()

        # SELF - Definitions
        self.middleware = add_middleware
        self.extensions = graphql_extensions
        self.apps = super_api
        self.types = app_models
        self.on_event = events_extensions

        # Command-Line-Interface
        self.cli = None
        if is_cli:
            from .commands import cli as core_cli

            # Core CLI
            all_clis.append(core_cli)

            # Start Click CLI
            def init_commands(extras: list = None):
                extras = extras or []
                click.clear()
                if extras:
                    all_clis.extend(extras)
                return command_collection(all_clis)()

            self.cli = init_commands



"""
    Strawberry - GraphQL (Connection, PageInfo, Edges)
"""

from dataclasses import field
from typing import Generic, TypeVar

import strawberry
from strawberry.scalars import JSON

GenericType = TypeVar("GenericType")


@strawberry.type
class Connection(Generic[GenericType]):
    """Represents a paginated relationship between two entities"""

    page_info: "PageInfo"
    edges: list["Edge[GenericType]"]


@strawberry.type
class PageInfo:
    """Pagination context to navigate objects with cursor-based pagination"""

    length: int
    pages: int
    extra: JSON = field(default_factory=dict)


@strawberry.type
class Edge(Generic[GenericType]):
    """An edge may contain additional information of the relationship. This is the trivial case"""

    node: GenericType
    cursor: str
